
object AccessSet {
  def apply(indexes: Traversable[Index]): AccessSet = {
    val set = indexes.toSet
    AccessSet(set, BoundingBox(set))
  }

  def apply(index: Index): AccessSet = {
    val set = Set(index)
    AccessSet(set, BoundingBox(set))
  }

  val center = AccessSet(List[Index](Index(0, 0, 0)))
  val empty = AccessSet(List[Index]())
}

case class AccessSet(indexes: Set[Index], boundingBox: BoundingBox) {
  // compute the minkowski sum
  def +(other: AccessSet) = {
    val combinations =
      for {
        thisIndex <- indexes
        otherIndex <- other.indexes
      } yield thisIndex + otherIndex
    AccessSet(combinations.toSet)
  }

  // compute union of two access sets
  def union(other: AccessSet) = {
    AccessSet(indexes union other.indexes)
  }

  // compute intersection of two access sets
  def intersect(other: AccessSet) = {
    AccessSet(indexes intersect other.indexes)
  }

  // compute the set difference
  def minus(other: AccessSet) = {
    AccessSet(indexes -- other.indexes)
  }

  // compute the element in the middle
  def middle = {
    indexes.map { index => (index, Math.abs(index.i) + Math.abs(index.j) + Math.abs(index.k)) }.minBy(_._2)._1
  }

  // compute the access set cardinality on a specific domain
  def cardinality(domain: Domain) = {
    // an empty access set has cardinality 0 otherwise perform a somewhat lengthy computation
    if(indexes.isEmpty) {
      0L // return cardinality 0 for an empty access set
    } else {
      // compute access set
      def computeAccessSet(domain: Domain) = {
        val interior = for {
          i <- 1 to domain.iSize
          j <- 1 to domain.jSize
          k <- 1 to domain.kSize
        } yield Index(i, j, k)
        this + AccessSet(interior)
      }
      // compute dwarf domain size
      val dwarfDomain = Domain(
        Math.min(1 + boundingBox.to.i - boundingBox.from.i, domain.iSize),
        Math.min(1 + boundingBox.to.j - boundingBox.from.j, domain.jSize),
        Math.min(1 + boundingBox.to.k - boundingBox.from.k, domain.kSize)
      )
      // compute a dwarf set
      val dwarfSet = computeAccessSet(dwarfDomain)

      // setup a dwarf domain and stretch it to the desired size
      val cardinality = DwarfDomain.stretchAndCount(dwarfSet, dwarfDomain, domain, boundingBox)
      //assert(cardinality == computeAccessSet(domain).indexes.size)
      // finally return the cardinality
      cardinality
    }
  }
  
  // create a domain that is extended by the bounding box
  def computeBoundingBoxDomain(domain: Domain) = {
    val iExtension = boundingBox.to.i - boundingBox.from.i
    val jExtension = boundingBox.to.j - boundingBox.from.j
    val kExtension = boundingBox.to.k - boundingBox.from.k
    Domain(domain.iSize + iExtension, domain.jSize + jExtension, domain.kSize + kExtension)
  }

  // convert the access set to an isl map
  def convertToISLMap = {
    // define the space of our mapping
    val space = isl.Space.alloc(ISLContext.ctx, 0, 3, 3)
    // add an index constraint for each index
    val accesses = indexes.toList.foldLeft(isl.Map.empty(space)) { (current, index) =>
      val dimensions = List(index.i, index.j, index.k)
      val constraints = ((0 to 2) zip dimensions).map { dimension =>
        val empty = isl.Constraint.equalityAlloc(isl.LocalSpace.fromSpace(space))
        val value = empty.setConstantVal(isl.Val.fromInt(ISLContext.ctx, dimension._2))
        val in = value.setCoefficientVal(isl.DimType.In, dimension._1, isl.Val.fromInt(ISLContext.ctx, 1))
        val out = in.setCoefficientVal(isl.DimType.Out, dimension._1, isl.Val.fromInt(ISLContext.ctx, -1))
        out
      }
      val indexes = constraints.foldLeft(isl.BasicMap.universe(space)) { (current, constraint) =>
        current.addConstraint(constraint)
      }
      current.union(indexes)
    }
    accesses.coalesce()
  }
  
  override def toString = indexes.mkString(", ")
}