/**
 * class holding the stencil group bandwidth information
 */
case class BandwidthRequirements(stencilGroup: StencilGroup,
                            computeOperations: List[Option[Long]],
                            verticalCommunication: List[Map[String, Long]], 
                            lateralCommunication: List[Map[String, List[Long]]],
                            nestedRequirements: List[Option[BandwidthRequirements]]) {
  // total communication requirements for the individual stencil group nodes
  val totalVerticalCommunication = verticalCommunication.map { communication => communication.values.sum }
  val totalLateralCommunication = lateralCommunication.map { communication =>
    if(communication.isEmpty) List.empty[Long]
    else communication.values.reduce { (l1, l2) => (l1 zip l2).map { case (e1, e2) => e1 + e2} }
  }

  // compute the bandwidth usage of a stencil group
  def computeBandwidthUsage(tilingHierarchy: List[Tiling]): List[BandwidthUsage] = {
    // compute the individual node bandwidth usages and extend them with the compute and
    stencilGroup.nodes.zipWithIndex.map {
      case (nestedStencil: Stencil, index) =>
        // estimate the execution time based on the computational intensity
        val communication = totalVerticalCommunication(index)
        val computation = computeOperations(index).get
        val executionTime = Math.max(
          computation.toDouble / stencilGroup.tiling.verticalMemory.computeThroughput,
          communication.toDouble / stencilGroup.tiling.verticalMemory.memoryBandwidth
        )
        val lateralMemories = (stencilGroup.tiling :: tilingHierarchy).map(_.lateralMemory).reverse
        val haloCommunication = lateralCommunication(index).getOrElse(nestedStencil.id, List.fill(lateralMemories.size)(0L))
        val haloExchangeTime = (lateralMemories zip haloCommunication).map { case (memory, communication) =>
          communication / memory.memoryBandwidth
        }.max
        List(
          BandwidthUsage(
            nestedStencil,
            Math.max(executionTime, haloExchangeTime),
            computation,
            Map(stencilGroup.tiling.verticalMemory.id -> communication),
            (lateralMemories.map(_.id) zip haloCommunication).toMap
          )
        )
      case (nestedStencilGroup: StencilGroup, index) =>
        // extract the bandwidth requirement of the nested node and compute the corresponding bandwidth requirements
        val nestedBandwidthUsage = nestedRequirements(index).get.computeBandwidthUsage(stencilGroup.tiling :: tilingHierarchy)
        // if both tiling hierarchy levels are on the same level we do not consider any vertical communication
        if(stencilGroup.tiling.verticalMemory.id == nestedStencilGroup.tiling.verticalMemory.id) {
          nestedBandwidthUsage
        } else {
          // distribute the communication among the nested bandwidth usage and a sentinel usage
          val communication = totalVerticalCommunication(index)
          val sentinel = BandwidthUsage(nestedStencilGroup, Double.MaxValue, 0L, Map.empty[String, Long], Map.empty[String, Long])
          val initialUsage = nestedBandwidthUsage ::: List(sentinel)
          val balancedUsage = initialUsage.scanLeft((sentinel, communication)) { (remainingCommunication, bandwidthUsage) =>
            bandwidthUsage.reserveBandwidth(stencilGroup.tiling.verticalMemory, remainingCommunication._2)
          }.tail.map(_._1)
          // adapt the execution time of the sentinel
          val id = stencilGroup.tiling.verticalMemory.id
          val remainingCommunication = balancedUsage.last.verticalCommunication(id)
          val executionTime = remainingCommunication.toDouble / stencilGroup.tiling.verticalMemory.memoryBandwidth
          if(executionTime > 0.0) {
            balancedUsage.dropRight(1) :::
              List(
                BandwidthUsage(
                  nestedStencilGroup,
                  executionTime, 0L,
                  Map(id -> remainingCommunication),
                  Map.empty[String, Long]
                )
              )
          } else {
            balancedUsage.dropRight(1)
          }
        }
    }.flatten
  }

  // lazy evaluation of bandwidth usage (can be used to extract execution time of a subset of the stencil group)
  lazy val bandwidthUsage = computeBandwidthUsage(List.empty[Tiling])

  // get the stencil group execution time
  def executionTime = bandwidthUsage.map(_.executionTime).sum

  // print the bandwidth analysis
  def printAnalysisInformation(prefix: String) = {
    // prepare analysis info
    val usage = computeBandwidthUsage(List.empty[Tiling])
    val executionTime = usage.map(_.executionTime).sum
    val nodeAnalysis = usage.map(_.printAnalysisInformation("\t" + prefix)).mkString("\n")
    s"$prefix $stencilGroup: execution time=${executionTime * 1000 * 1000} us\n $nodeAnalysis"
  }

  // pretty print information
  def printNodeInformation(prefix: String) = {
    s"$prefix $stencilGroup: \n" + prettyPrint(s"\t$prefix")
  }

  // pretty print information
  def prettyPrint(prefix: String): String = {
    val computation = computeOperations.map { case Some(intensity) => s"flops=$intensity"; case None => "flops=" }
    val vertical = totalVerticalCommunication.map { case communication => s"vertical=$communication"}
    val lateral = totalLateralCommunication.map { case communication => s"lateral=${communication.mkString(">")}"}
    val nested = nestedRequirements.map { case Some(requirement) => s"\n${requirement.prettyPrint("\t" + prefix)}"; case None => "" }
    val information = for(i <- 0 until stencilGroup.nodes.size) yield(s" ${stencilGroup.nodes(i).id}: ${computation(i)} ${vertical(i)} ${lateral(i)} ${nested(i)}")
    s"$prefix" + information.mkString(s"\n${prefix}")
  }

  override def toString = s"-> $stencilGroup: \n" + prettyPrint("\t->")
}
