/**
 * trait defining the optimizer interface
 */
trait Optimizer {
  // helper method
  def pickCandidate(candidate1: StencilGroup, candidate2: StencilGroup) = {
    val (fastCandidate, slowCandidate) = if(candidate1.executionTime < candidate2.executionTime) (candidate1, candidate2) else (candidate2, candidate1)
    if(!fastCandidate.isFeasible && slowCandidate.isFeasible) slowCandidate
    else fastCandidate
  }

  // optimization interface
  def searchOptimalTiling(stencils: List[Stencil], outputs: List[String], searchConfiguration: List[(Tiling, Boolean)]): StencilGroup

  // additional optimizer that iterates over multiple topological orders
  def searchOptimalTopology(range: (Int, Int), stencils: List[Stencil], outputs: List[String], searchConfiguration: List[(Tiling, Boolean)]) = {
    // method that computes all topological orders
    def computeTopologicalOrders(placed: List[Stencil], free: List[Stencil]): List[List[Stencil]] = {
      if (free.isEmpty) {
        List(placed)
      } else {
        // find all variable stencils that have no dependencies to all other stencils
        val stencils = free.filter { stencil =>
          val blackList = free.map(_.id).toSet
          val dependencies = stencil.dataDependencies.keySet
          (blackList intersect dependencies).isEmpty
        }
        // iterate over all candidates
        stencils.foldLeft(List.empty[List[Stencil]]) { (candidates, stencil) =>
          // prepare placed and free lists
          val freeStencils = free.filter(free => free.id != stencil.id)
          val placedStencils = placed ::: List(stencil)
          candidates ::: computeTopologicalOrders(placedStencils, freeStencils)
        }
      }
    }
    // compute all orders and search the specified range
    println(s"-> searching all topological orders in range (${range._1}, ${range._2})")
    val topologicalOrders = computeTopologicalOrders(List.empty[Stencil], stencils).slice(range._1, range._2)
    topologicalOrders.foldLeft(StencilGroup(searchConfiguration.head._1, List.empty[Node], List.empty[String])) { (optimum, stencils) =>
      val candidate = searchOptimalTiling(stencils, outputs, searchConfiguration)
      println(s"-> $candidate: execution time=${candidate.executionTime * 1000 * 1000} us")
      // regularly run garbage collection
      System.gc()
      Thread.sleep(200)
      pickCandidate(optimum, candidate)
    }
  }
}
