/**
 * class running an exhaustive search for finding the optimal implementation
 */
class OptimizerBF extends Optimizer {

  // performs a brute force search
  def searchOptimalTiling(stencils: List[Stencil], outputs: List[String], searchConfiguration: List[(Tiling, Boolean)]) = {
    var counter = 0
    // compute sets that hold the results respectively dependencies up to respectively after the given index
    val results = stencils.scanLeft(Set.empty[String]) { (currentResults, stencil) =>
      currentResults + stencil.id
    }.toVector
    val dependencies = stencils.scanRight(outputs.toSet) { (stencil, currentDependencies) =>
      currentDependencies ++ stencil.dataDependencies.keys.toSet
    }.toVector
    // implement the actual search logic
    def searchOptimalTilingImpl(nodes: List[Node], fromIndexes: IndexedSeq[Int], toIndexes: IndexedSeq[Int], searchConfiguration: List[(Tiling, Boolean)]): StencilGroup = {
      val (tiling, search) = searchConfiguration.head
      if(searchConfiguration.tail.isEmpty) {
        // create the bottom level stencil group
        val outputs = (results(toIndexes.last) -- results(fromIndexes.head)) intersect dependencies(toIndexes.last)
        val candidate = StencilGroup(tiling, nodes, outputs.toList)
        counter += 1
        /* println(s"candidate $counter: $candidate => estimated execution time: ${candidate.executionTime}") */
        System.gc()
        Thread.sleep(100)
        candidate
      } else {
        if(!search) {
          // perform no optimization on this level
          val forwardNodes = nodes.map { group => StencilGroup(tiling, List(group), group.outputs) }
          searchOptimalTilingImpl(forwardNodes, fromIndexes, toIndexes, searchConfiguration.tail)
        } else {
          assert(nodes.nonEmpty)
          // split the level in arbitrary groups and find the optimal combination
          val size = nodes.size - 1
          def computeCandidate(candidate: StencilGroup, counter: Int) = {
            // compute from and to index sequences
            val start = (0 until size).scanLeft(0) { (current, index) => if (((1 << index) & counter) == 0) current else index + 1}.distinct
            val end = (0 until size).scanRight(size + 1) { (index, current) => if (((1 << index) & counter) == 0) current else index + 1}.distinct
            val from = (0 until size).scanLeft(fromIndexes.head) { (current, index) => if (((1 << index) & counter) == 0) current else fromIndexes(index + 1)}.distinct
            val to = (0 until size).scanRight(toIndexes.last) { (index, current) => if (((1 << index) & counter) == 0) current else toIndexes(index)}.distinct
            // create a new stencil group for each distinct index pair
            val stencilGroups = ((start zip end) zip (from zip to)).map { range =>
              val start = range._1._1
              val end = range._1._2
              val from = range._2._1
              val to = range._2._2
              val outputs = (results(to) -- results(from)) intersect dependencies(to)
              StencilGroup(tiling, nodes.slice(start, end).toList, outputs.toList)
            }
            pickCandidate(candidate, searchOptimalTilingImpl(stencilGroups.toList, from, to, searchConfiguration.tail))
          }
          // run the actual search starting with an infeasible sentinel stencil group
          (0 until (1 << size)).aggregate(StencilGroup(tiling, List.empty[Node], List.empty[String])) (computeCandidate, pickCandidate)
        }
      }
    }
    // start the search
    val fromIndexes = 0 until stencils.size
    val toIndexes = 1 until stencils.size + 1
    searchOptimalTilingImpl(stencils, fromIndexes, toIndexes, searchConfiguration)
  }
}
