/**
 * class holding the stencil group storage information
 */
case class StorageRequirements(stencilGroup: StencilGroup,
                               temporaryStorage: Map[String, Long],
                               nestedRequirements: List[Option[StorageRequirements]]) {
  // total storage requirements
  val totalStorageRequirements = temporaryStorage.values.sum
  
  // compute group wise storage usage on all memory levels
  def computeStorageRequirementImpl(parentRequirements: Map[String, (Long, Long)]): List[Map[String, (Long, Long)]] = {
    // update the parent requirements
    val id = stencilGroup.tiling.verticalMemory.id
    val updatedRequirements = parentRequirements.updated(id,
      (
        parentRequirements.getOrElse(id, (0L, 0L))._1 + totalStorageRequirements,
        stencilGroup.tiling.verticalMemory.memorySize
      )
    )
    // recursively compute the child storage requirements
    if (nestedRequirements.flatten.isEmpty) {
      List(updatedRequirements)
    } else {
      nestedRequirements.flatten.map { requirement =>
        requirement.computeStorageRequirementImpl(updatedRequirements)
      }.flatten
    }
  }

  // compute the storage requirements of the full program
  def computeStorageRequirement = { computeStorageRequirementImpl(Map.empty[String, (Long, Long)]) }

  // verify the requirements
  def verifyStorageRequirements = {
    val requirements = computeStorageRequirementImpl(Map.empty[String, (Long, Long)]).flatMap(_.toSeq)
    val ok = requirements.map { case (id, (usage, limit)) => if (usage <= limit) true else false }
    !ok.contains(false)
  }

  // pretty print information
  def prettyPrint(prefix: String): String = {
    val header = s"$prefix $stencilGroup: storage=$totalStorageRequirements"
    val nested = nestedRequirements.flatten.map { requirement => s"${requirement.prettyPrint("\t" + prefix)}" }
    (header :: nested).mkString("\n")
  }

  override def toString = prettyPrint("->")
}
