import org.scalatest._

/**
 * test the storage computation
 */
class EvaluateStorageSpec extends FlatSpec with Matchers {
  val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1, 0, 0), Index(1, 0, 0), Index(0, -1, 0), Index(0, 1, 0), Index(0, 0, 0)))))
  val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1, 0, 0), Index(0, 0, 0)))))
  val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0, 1, 0), Index(0, 0, 0)))))
  val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1, 0, 0), Index(0, 0, 0))), "flj" -> AccessSet(List(Index(0, -1, 0), Index(0, 0, 0)))))

  "a StencilGroup" should "consider input and temporary fields for the storage evaluation" in {
    val domain = Domain(32, 32, 32)
    val tiling = Tiling(0, Memory.DeviceMemory, Memory.DeviceMemory, domain, OF)

    val stencilGroup = StencilGroup(tiling, List(lap, fli, flj, out), List("out"))

    // compute on the fly storage requirements
    val storageIn = domain.cardinality + 8 * 32 * 32 + 4 * 32
    val storageLap = domain.cardinality + 4 * 32 * 32
    val storageFli = domain.cardinality + 1 * 32 * 32
    val storageFlj = domain.cardinality + 1 * 32 * 32
    val storageWgt = domain.cardinality

    val storageRequirement = stencilGroup.evaluateStorage

    storageRequirement.stencilGroup.tiling.verticalMemory should equal (Memory.DeviceMemory)
    storageRequirement.totalStorageRequirements should equal (storageIn + storageLap + storageFli + storageFlj + storageWgt)
  }

  "a StencilGroup" should "consider nested stencil groups for the storage evaluation" in {
    val domain1 = Domain(32, 32, 32)
    val domain2 = Domain(8, 8, 8)
    val tiling1 = Tiling(0, Memory.DeviceMemory, Memory.DeviceMemory, domain1, OF)
    val tiling2 = Tiling(1, Memory.SharedMemory, Memory.DeviceMemory, domain2, HEPar)

    val stencilGroup2 = StencilGroup(tiling2, List(fli, flj, out), List("out"))
    val stencilGroup1 = StencilGroup(tiling1, List(lap, stencilGroup2), List("out"))

    val storageRequirement1 = stencilGroup1.evaluateStorage

    // compute on-the-fly storage requirements
    {
      val storageIn = domain1.cardinality + 8 * 32 * 32 + 4 * 32
      val storageLap = domain1.cardinality + 4 * 32 * 32 // we do not consider the outer boundary!
      val storageOut = domain1.cardinality

      storageRequirement1.stencilGroup.tiling.verticalMemory should equal(Memory.DeviceMemory)
      storageRequirement1.totalStorageRequirements should equal(storageIn + storageLap + storageOut)
    }

    // verify nested level with halo exchange
    {
      val storageIn = domain2.cardinality
      val storageWgt = domain2.cardinality
      val storageLap = domain2.cardinality + 2 * 8 * 8
      val storageFli = domain2.cardinality + 1 * 8 * 8
      val storageFlj = domain2.cardinality + 1 * 8 * 8

      storageRequirement1.nestedRequirements.flatten.head.stencilGroup.tiling.verticalMemory should equal(Memory.SharedMemory)
      storageRequirement1.nestedRequirements.flatten.head.totalStorageRequirements should equal(storageIn + storageWgt + storageFli + storageFlj + storageLap)
    }
  }
}
